from django import forms
from edc_crf.modelform_mixins import CrfModelFormMixin, CrfSingletonModelFormMixin

from edc_he.form_validators.assets_form_validator import (
    HealthEconomicsAssetsFormValidator,
)
from edc_he.forms import HealthEconomicsModelFormMixin

from ..models import HealthEconomicsAssets


class HealthEconomicsAssetsForm(
    CrfSingletonModelFormMixin,
    HealthEconomicsModelFormMixin,
    CrfModelFormMixin,
    forms.ModelForm,
):
    form_validator_cls = HealthEconomicsAssetsFormValidator

    def clean(self):
        self.raise_if_singleton_exists()
        return super().clean()

    class Meta:
        model = HealthEconomicsAssets
        fields = "__all__"



from decimal import Decimal

from django.test import TestCase

from edc_he.calculators import (
    InvalidAreaUnitsError,
    acres_to_sq_meters,
    convert_to_sq_meters,
    decimals_to_sq_meters,
    hectares_to_sq_meters,
    sq_feet_to_sq_meters,
)
from edc_he.constants import ACRES, DECIMALS, HECTARES, SQ_FEET, SQ_METERS


class TestCalculators(TestCase):
    def test_acres_to_sq_metres(self):
        for area, expected_output in [
            (1, Decimal("4046.8564224")),
            (2, Decimal("8093.7128448")),
            (10, Decimal("40468.564224")),
        ]:
            with self.subTest(area=area, expected_output=expected_output):
                self.assertEqual(acres_to_sq_meters(area), expected_output)

    def test_decimals_to_sq_metres(self):
        for area, expected_output in [
            (1, Decimal("40.468564224")),
            (2, Decimal("80.937128448")),
            (5, Decimal("202.34282112")),
            (10, Decimal("404.68564224")),
            (12.5, Decimal("505.8570528")),
            (20, Decimal("809.37128448")),
            (50, Decimal("2023.4282112")),
            (99, Decimal("4006.387858176")),
            (100, Decimal("4046.8564224")),
        ]:
            with self.subTest(area=area, expected_output=expected_output):
                self.assertEqual(decimals_to_sq_meters(area), expected_output)

    def test_hectares_to_sq_metres(self):
        for area, expected_output in [
            (1, Decimal("10_000")),
            (2, Decimal("20_000")),
            (5, Decimal("50_000")),
            (10, Decimal("100_000")),
        ]:
            with self.subTest(area=area, expected_output=expected_output):
                self.assertEqual(hectares_to_sq_meters(area), expected_output)

    def test_sq_feet_to_sq_metres(self):
        for area, expected_output in [
            (50 * 50, Decimal("232.2576")),
            (100 * 100, Decimal("929.0304")),
            (100 * 50, Decimal("464.5152")),
        ]:
            with self.subTest(area=area, expected_output=expected_output):
                self.assertEqual(sq_feet_to_sq_meters(area), expected_output)

    def test_equivalent_conversions(self):
        # 1 hectare == 10,000 square meters
        self.assertEqual(hectares_to_sq_meters(1), 10_000)

        # 1 hectare ~= 2.47 acres (within 5 square meters)
        self.assertAlmostEqual(hectares_to_sq_meters(1), acres_to_sq_meters(2.47), delta=5)

        # 259 hectares ~= 640 acres (within 12 square meters)
        self.assertAlmostEqual(hectares_to_sq_meters(259), acres_to_sq_meters(640), delta=12)

        # 1 acre ~= 4046.86 square meters (within 1 square cm)
        self.assertAlmostEqual(acres_to_sq_meters(1), Decimal(4046.86), delta=0.01)

        # 1 acre == 43,560 square feet
        self.assertEqual(acres_to_sq_meters(1), sq_feet_to_sq_meters(43_560))

        # 1 acre == 100 decimals
        self.assertEqual(acres_to_sq_meters(1), decimals_to_sq_meters(100))

        # 50 decimals == half an acre
        self.assertEqual(decimals_to_sq_meters(50), acres_to_sq_meters(1) / 2)
        self.assertEqual(decimals_to_sq_meters(50), acres_to_sq_meters(0.5))

        # 25 decimals == quarter an acre
        self.assertEqual(decimals_to_sq_meters(25), acres_to_sq_meters(1) / 4)
        self.assertEqual(decimals_to_sq_meters(25), acres_to_sq_meters(0.25))

        # 12.5 decimals == eighth an acre
        self.assertEqual(decimals_to_sq_meters(12.5), acres_to_sq_meters(1) / 8)
        self.assertEqual(decimals_to_sq_meters(12.5), acres_to_sq_meters(0.125))

        # 1 square meter ~= 10.7 square feet
        self.assertAlmostEqual(Decimal("1"), sq_feet_to_sq_meters(10.7639), delta=0.001)

        # 1 foot == 0.348 meters
        # 1 square foot == 0.3048 * 0.3048 square meters
        self.assertEqual(sq_feet_to_sq_meters(1), Decimal("0.3048") * Decimal("0.3048"))

    def test_convert_to_sq_meters(self):
        self.assertEqual(convert_to_sq_meters(area=123, area_units=SQ_METERS), Decimal("123"))

        self.assertEqual(
            convert_to_sq_meters(area=1, area_units=ACRES),
            Decimal("4046.8564224"),
        )

        self.assertEqual(
            convert_to_sq_meters(area=100, area_units=DECIMALS),
            Decimal("4046.8564224"),
        )

        self.assertEqual(convert_to_sq_meters(area=1, area_units=HECTARES), Decimal("10_000"))

        self.assertEqual(
            convert_to_sq_meters(area=200 * 200, area_units=SQ_FEET), Decimal("3716.1216")
        )

    def test_convert_to_sq_meters_unrecognised_units_raises(self):
        with self.assertRaises(InvalidAreaUnitsError):
            convert_to_sq_meters(area=2, area_units="unsupported_units")



from faker import Faker
from model_bakery.recipe import Recipe

from .tests.models import HealthEconomicsAssets, HealthEconomicsProperty

fake = Faker()

healtheconomicsassets = Recipe(HealthEconomicsAssets, subject_visit=None)
healtheconomicsproperty = Recipe(HealthEconomicsProperty, subject_visit=None)



from django.contrib import admin
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from django_audit_fields.admin import audit_fieldset_tuple
from edc_crf.admin import crf_status_fieldset_tuple


def get_income_fieldsets_tuple() -> list[tuple]:
    fieldsets = []
    sources = [
        ("wages", _("Wages / Salary")),
        ("selling", _("Selling, trading or hawking")),
        ("rental_income", _("Rental income")),
        ("pension", _("Pension, etc.")),
        ("ngo_assistance", _("NGO assistance")),
        ("interest", _("Interest and dividends")),
        ("internal_remit", _("Money transfers (domestic)")),
        ("external_remit", _("Remittances (from outside the country)")),
        ("more_sources", _("Other sources of income")),
    ]
    for fld, label in sources:
        fields = [
            fld,
            f"{fld}_value_known",
            f"{fld}_value",
        ]
        if fld == "external_remit":
            fields.extend(["external_remit_currency", "external_remit_currency_other"])
        if fld == "more_sources":
            fields.insert(1, "more_sources_other")

        fieldsets.append(
            (
                label,
                {
                    "description": format_html(
                        _(
                            "Estimate the total amount of income from this source "
                            "for the <b>household</b> over the time period indicated"
                        )
                    ),
                    "fields": tuple(fields),
                },
            )
        )
    return fieldsets


class HealthEconomicsIncomeModelAdminMixin:
    form = None

    additional_instructions = format_html(
        _(
            "<H5><B><font color='orange'>Interviewer to read</font></B></H5>"
            "<P>We want to learn about the household and we use these questions "
            "to get an understanding of wealth and opportunities in the community.</P>"
            "<p>Now, I will ask about <b>income for the household</b> from paid work or "
            "other sources.</p><p>I know it may be difficult to calculate those figures, "
            "but please do try to give amounts as accurately as possible. Remember "
            "that <b>all information will be kept strictly confidential</b>. This "
            "information is important to assess overall health and well-being of "
            "people in your household, compared to other similar households.</p><p>"
            "I am now going to read you a list of possible sources of income. "
            "Thinking over the last 12 months, can you tell me what the average "
            "earnings of the household have been per week or per month or per year? "
            "Please tell me whichever time period that is easier for you.</p>"
        )
    )
    fieldsets = (
        (None, {"fields": ("subject_visit", "report_datetime")}),
        *get_income_fieldsets_tuple(),
        (
            "Dependents, financial status",
            {
                "fields": (
                    "external_dependents",
                    "income_enough",
                    "financial_status",
                    "financial_status_compare",
                )
            },
        ),
        (
            "Debt / Loans",
            {
                "description": format_html(
                    _(
                        "<H5><B><font color='orange'>Interviewer to read</font></B></H5>"
                        "<p>Now, I will ask about the total current debt/loans for the "
                        "household. I know it may be difficult to calculate this, but please "
                        "do try to give amounts as accurately as possible. Remember that "
                        "all information will be kept strictly confidential. <B>It is not "
                        "possible to offer any help for your debts through this project</B>, "
                        "but this information will contribute to research and is important "
                        "to assess the overall health and well-being of people in your "
                        "household, compared to other similar households.</p>"
                    )
                ),
                "fields": (
                    "household_debt",
                    "household_debt_value",
                ),
            },
        ),
        crf_status_fieldset_tuple,
        audit_fieldset_tuple,
    )

    radio_fields = {
        "wages": admin.VERTICAL,
        "wages_value_known": admin.VERTICAL,
        "selling": admin.VERTICAL,
        "selling_value_known": admin.VERTICAL,
        "rental_income": admin.VERTICAL,
        "rental_income_value_known": admin.VERTICAL,
        "pension": admin.VERTICAL,
        "pension_value_known": admin.VERTICAL,
        "ngo_assistance": admin.VERTICAL,
        "ngo_assistance_value_known": admin.VERTICAL,
        "interest": admin.VERTICAL,
        "interest_value_known": admin.VERTICAL,
        "internal_remit": admin.VERTICAL,
        "internal_remit_value_known": admin.VERTICAL,
        "external_remit": admin.VERTICAL,
        "external_remit_value_known": admin.VERTICAL,
        "external_remit_currency": admin.VERTICAL,
        "more_sources": admin.VERTICAL,
        "more_sources_value_known": admin.VERTICAL,
        "income_enough": admin.VERTICAL,
        "financial_status": admin.VERTICAL,
        "financial_status_compare": admin.VERTICAL,
        "household_debt": admin.VERTICAL,
        "crf_status": admin.VERTICAL,
    }



from .assets_form import HealthEconomicsAssetsForm
from .health_economics_form import HealthEconomicsForm
from .household_head import HealthEconomicsHouseholdHeadForm
from .income_form import HealthEconomicsIncomeForm
from .patient_form import HealthEconomicsPatientForm
from .property_form import HealthEconomicsPropertyForm
